import importlib.util
import os
from typing import Union, Optional, TYPE_CHECKING, cast

from ibind import var
from ibind.base.rest_client import RestClient, Result
from ibind.client.ibkr_client_mixins.accounts_mixin import AccountsMixin
from ibind.client.ibkr_client_mixins.contract_mixin import ContractMixin
from ibind.client.ibkr_client_mixins.marketdata_mixin import MarketdataMixin
from ibind.client.ibkr_client_mixins.order_mixin import OrderMixin
from ibind.client.ibkr_client_mixins.portfolio_mixin import PortfolioMixin
from ibind.client.ibkr_client_mixins.scanner_mixin import ScannerMixin
from ibind.client.ibkr_client_mixins.session_mixin import SessionMixin
from ibind.client.ibkr_client_mixins.watchlist_mixin import WatchlistMixin
from ibind.client.ibkr_utils import Tickler
from ibind.support.errors import ExternalBrokerError
from ibind.support.logs import new_daily_rotating_file_handler, project_logger

if TYPE_CHECKING:  # pragma: no cover
    from ibind.oauth import OAuthConfig

_LOGGER = project_logger(__file__)


class IbkrClient(RestClient, AccountsMixin, ContractMixin, MarketdataMixin, OrderMixin, PortfolioMixin, ScannerMixin, SessionMixin, WatchlistMixin):
    """
    A client class for interfacing with the IBKR API, extending the RestClient class.

    This subclass of RestClient is specifically designed for the IBKR API. It inherits
    the foundational REST API interaction capabilities from RestClient and adds functionalities
    particular to the IBKR API, such as specific endpoint handling.

    The class provides methods to perform various operations with the IBKR API, such as
    fetching stock data, submitting orders, and managing account information.

    See: https://interactivebrokers.github.io/cpwebapi/endpoints

    Note:
        - All endpoint mappings are defined as class mixins, categorised similar to the IBKR REST API documentation. See appropriate mixins for more information.
    """

    def __init__(
            self,
            account_id: Optional[str] = var.IBIND_ACCOUNT_ID,
            url: str = var.IBIND_REST_URL,
            host: str = '127.0.0.1',
            port: str = '5000',
            base_route: str = '/v1/api/',
            cacert: Union[str, os.PathLike, bool] = var.IBIND_CACERT,
            timeout: float = 10,
            max_retries: int = 3,
            use_session: bool = var.IBIND_USE_SESSION,
            auto_register_shutdown: bool = var.IBIND_AUTO_REGISTER_SHUTDOWN,
            use_oauth: bool = var.IBIND_USE_OAUTH,
            oauth_config: 'OAuthConfig' = None,

    ) -> None:
        """
        Parameters:
            account_id (str): An identifier for the account. Defaults to None.
            url (str): The base URL for the REST API. Defaults to None.
                       If 'use_oauth' is specified, the url is taken from oauth_config.
                       Only if it couldn't be found in oauth_config, this url is used,
                       or the parameters host, port and base_route.
            host (str, optional): Host for the IBKR REST API. Defaults to 'localhost'.
            port (str, optional): Port for the IBKR REST API. Defaults to '5000'
            base_route (str, optional): Base route for the IBKR REST API. Defaults to '/v1/api/'.
            cacert (Union[os.PathLike, bool], optional): Path to the CA certificate file for SSL verification,
                                                         or False to disable SSL verification. Always True when
                                                         use_oauth is True. Defaults to False.
            timeout (float, optional): Timeout in seconds for the API requests. Defaults to 10.
            max_retries (int, optional): Maximum number of retries for failed API requests. Defaults to 3.
            use_session (bool, optional): Whether to use a persistent session for making requests. Defaults to True.
            auto_register_shutdown (bool, optional): Whether to automatically register a shutdown handler for this client. Defaults to True.
            use_oauth (bool, optional): Whether to use OAuth authentication. Defaults to False.
            oauth_config (OAuthConfig, optional): The configuration for the OAuth authentication. OAuth1aConfig is used if not specified.
        """

        self._use_oauth = use_oauth

        if self._use_oauth:
            from ibind.oauth.oauth1a import OAuth1aConfig
            # cast to OAuth1aConfig for type checking, since currently 1.0a is the only version used
            self.oauth_config = cast(OAuth1aConfig, oauth_config) if oauth_config is not None else OAuth1aConfig()
            url = url if url is not None and self.oauth_config.oauth_rest_url is None else self.oauth_config.oauth_rest_url

        if url is None:
            url = f'https://{host}:{port}{base_route}'

        self.account_id = account_id

        cacert = True if self._use_oauth else cacert
        super().__init__(
            url=url,
            cacert=cacert,
            timeout=timeout,
            max_retries=max_retries,
            use_session=use_session,
            auto_register_shutdown=auto_register_shutdown,
        )

        self.logger.info('#################')
        self.logger.info(f'New IbkrClient(base_url={self.base_url!r}, account_id={self.account_id!r}, ssl={self.cacert!r}, timeout={self._timeout}, max_retries={self._max_retries}, use_oauth={self._use_oauth})')

        if self._use_oauth:
            self.oauth_config.verify_config()

            if self.oauth_config.init_oauth:
                self.oauth_init(
                    maintain_oauth=self.oauth_config.maintain_oauth,
                    init_brokerage_session=self.oauth_config.init_brokerage_session,
                )

    def _make_logger(self):
        self._logger = new_daily_rotating_file_handler('IbkrClient', os.path.join(var.LOGS_DIR, f'ibkr_client_{self.account_id}'))

    def _request(
            self,
            method: str,
            endpoint: str,
            base_url: str = None,
            extra_headers: dict = None,
            log: bool = True,
            **kwargs
    ) -> Result:
        """ Handle IBKR-specific errors."""

        try:
            return super()._request(method, endpoint, base_url, extra_headers, log, **kwargs)
        except ExternalBrokerError as e:
            if 'Bad Request: no bridge' in str(e) and e.status_code == 400:
                raise ExternalBrokerError(f'IBKR returned 400 Bad Request: no bridge. Try calling `initialize_brokerage_session()` first.') from e
            raise

    def _get_headers(self, request_method: str, request_url: str):
        if (not self._use_oauth) or request_url == f'{self.base_url}{self.oauth_config.live_session_token_endpoint}':
            # No need for extra headers if we don't use oauth or getting live session token
            return {}

        # get headers for endpoints other than live session token request
        from ibind.oauth.oauth1a import generate_oauth_headers
        headers = generate_oauth_headers(
            oauth_config=self.oauth_config,
            request_method=request_method,
            request_url=request_url,
            live_session_token=self.live_session_token
        )

        return headers

    def generate_live_session_token(self):
        """
        Generates a new live session token for OAuth 1.0a authentication.

        This method requests a new OAuth live session token from the IBKR API using the configured
        OAuth credentials. The token is stored along with its expiration time and signature.

        The live session token is required for authenticated requests to IBKR's OAuth 1.0a API.

        Raises:
            ExternalBrokerError: If the token request fails.
        """
        from ibind.oauth.oauth1a import req_live_session_token
        self.live_session_token, self.live_session_token_expires_ms, self.live_session_token_signature \
            = req_live_session_token(self, self.oauth_config)

    def oauth_init(
            self,
            maintain_oauth: bool,
            init_brokerage_session: bool
    ):
        """
        Initializes the OAuth authentication flow for the IBKR API.

        This method sets up OAuth authentication by generating a live session token, validating it,
        and optionally starting a tickler to maintain the session. It also allows initializing a brokerage session if specified.

        OAuth authentication is required for certain IBKR API operations. The process includes:
        - Checking for the necessary cryptographic dependencies.
        - Generating and validating a live session token.
        - Optionally maintaining the session by running a tickler.
        - Initializing a brokerage session if required.

        Parameters:
            maintain_oauth (bool): If True, starts the Tickler process to keep the session alive.
            init_brokerage_session (bool): If True, initializes the brokerage session after authentication.

        Raises:
            ImportError: If the required cryptographic dependencies (`Crypto` module) are missing.
            RuntimeError: If live session token validation fails.

        See:
            - `generate_live_session_token`: Generates a new OAuth session token.
            - `validate_live_session_token`: Validates the generated OAuth session token.
            - `start_tickler`: Maintains the session by periodically sending requests.
            - `initialize_brokerage_session`: Establishes a brokerage session post-authentication.
        """
        _LOGGER.info(f'{self}: Initialising OAuth {self.oauth_config.version()}')

        if importlib.util.find_spec('Crypto') is None:
            raise ImportError('Installation lacks OAuth support. Please install by using `pip install ibind[oauth]`')

        # get live session token for OAuth authentication
        self.generate_live_session_token()

        # validate the live session token once
        from ibind.oauth.oauth1a import validate_live_session_token
        success = validate_live_session_token(
            live_session_token=self.live_session_token,
            live_session_token_signature=self.live_session_token_signature,
            consumer_key=self.oauth_config.consumer_key
        )
        if not success:
            raise RuntimeError("Live session token validation failed.")

        if maintain_oauth:
            self.start_tickler()

        if init_brokerage_session:
            self.initialize_brokerage_session()

    def start_tickler(self):
        """
        Starts the `Tickler` instance and starts it in a separate thread to maintain the OAuth session.

        The Tickler sends periodic requests to the IBKR API to prevent the session from expiring.
        This is necessary when using OAuth authentication to keep the connection active.

        Note:
            - The Tickler should be stopped when the session is no longer needed using `stop_tickler()`.

        """
        _LOGGER.info(f'{self}: Starting Tickler to maintain the connection alive')
        self._tickler = Tickler(self)
        self._tickler.start()

    def stop_tickler(self):
        """
        Stops the Tickler thread if the Tickler is running.

        The Tickler is responsible for maintaining an active session by sending periodic requests to
        the IBKR API. This method stops the Tickler process, preventing further requests.
        """
        if hasattr(self, '_tickler') and self._tickler is not None:
            self._tickler.stop()

    def close(self):
        if self._use_oauth and self.oauth_config.shutdown_oauth:
            self.oauth_shutdown()
        super().close()

    def oauth_shutdown(self):
        """
        Shuts down the OAuth session and cleans up resources.

        This method stops the Tickler process, which keeps the session alive, and logs out from
        the IBKR API to ensure a clean session termination.
        """
        _LOGGER.info(f'{self}: Shutting down OAuth')
        self.stop_tickler()
        self.logout()



from typing import TYPE_CHECKING, List, Dict

from ibind.base.rest_client import Result
from ibind.support.logs import project_logger
from ibind.support.py_utils import params_dict

if TYPE_CHECKING:  # pragma: no cover
    from ibind import IbkrClient

_LOGGER = project_logger(__file__)


class ScannerMixin():  # pragma: no cover
    """
    https://ibkrcampus.com/ibkr-api-page/cpapi-v1/#scanner
    """

    def scanner_parameters(self: 'IbkrClient') -> Result:
        """
        Returns an xml file containing all available parameters to be sent for the Iserver scanner request.
        """
        return self.get('iserver/scanner/params')

    def market_scanner(
            self: 'IbkrClient',
            instrument: str,
            type: str,
            location: str,
            filter: List[Dict[str, str]] = None
    ) -> Result:
        """
        Searches for contracts according to the filters specified in /iserver/scanner/params endpoint.
        Users can receive a maximum of 50 contracts from 1 request.

        Parameters:
            instrument (str): Instrument type as the target of the market scanner request. Found in the “instrument_list” section of the /iserver/scanner/params response.
            type (str): Scanner value the market scanner is sorted by. Based on the “scan_type_list” section of the /iserver/scanner/params response.
            location (str): Location value the market scanner is searching through. Based on the “location_tree” section of the /iserver/scanner/params response.
            filter (List[Dict[str, str]]): Contains any additional filters that should apply to response. Each filter object may include:
                - code (str): Code value of the filter. Based on the “code” value within the “filter_list” section of the /iserver/scanner/params response.
                - value (int): Value corresponding to the input for “code”.
        """
        params = params_dict(
            {
                'instrument': instrument,
                'type': type,
                'location': location,
            }, optional={'filter': filter or []}
        )
        return self.post('iserver/scanner/run', params)

    def hmds_scanner_parameters(self: 'IbkrClient') -> Result:
        """
        Query the parameter list for the HMDS market scanner.
        """
        return self.get('hmds/scanner/params')

    def hmds_market_scanner(
            self: 'IbkrClient',
            instrument: str,
            location: str,
            scan_code: str,
            sec_type: str,
            filter: List[Dict[str, str]],
            max_items: int = None,
    ) -> Result:
        """
        Request a market scanner from our HMDS service.
        Can return a maximum of 250 contracts.

        Parameters:
            instrument (str): Specify the type of instrument for the request. Found under the “instrument_list” value of the /hmds/scanner/params request.
            locations (str): Specify the type of location for the request. Found under the “location_tree” value of the /hmds/scanner/params request.
            scanCode (str): Specify the scanner type for the request. Found under the “scan_type_list” value of the /hmds/scanner/params request.
            secType (str): Specify the type of security type for the request. Found under the “location_tree” value of the /hmds/scanner/params request.
            filters (List[Dict[str, str]]): Array of objects containing all filters upon the scanner request. While “filters” must be specified in the body, no content in the array needs to be passed.
            maxItems (int, optional): Specify how many items should be returned. Default and maximum set to 250.
        """
        params = params_dict(
            {
                'instrument': instrument,
                'location': location,
                'scanCode': scan_code,
                'secType': sec_type
            }, optional={
                'maxItems': max_items,
                'filter': filter
            }
        )
        return self.post('hmds/scanner/run', params)



import datetime
from typing import Union, TYPE_CHECKING, List, Dict

from ibind.base.rest_client import Result
from ibind.client.ibkr_definitions import snapshot_by_id
from ibind.client.ibkr_utils import StockQuery, StockQueries, cleanup_market_history_responses
from ibind.support.errors import ExternalBrokerError
from ibind.support.logs import project_logger
from ibind.support.py_utils import ensure_list_arg, OneOrMany, execute_in_parallel, params_dict

if TYPE_CHECKING:  # pragma: no cover
    from ibind import IbkrClient

_LOGGER = project_logger(__file__)


class MarketdataMixin():
    """
    https://ibkrcampus.com/ibkr-api-page/cpapi-v1/#md
    """

    @ensure_list_arg('conids', 'fields')
    def live_marketdata_snapshot(self: 'IbkrClient', conids: OneOrMany[str], fields: OneOrMany[str]) -> Result:  # pragma: no cover
        """
        Get Market Data for the given conid(s).

        A pre-flight request must be made prior to ever receiving data.

        Parameters:
            conids (OneOrMany[str]): Contract identifier(s) for the contract of interest.
            fields (OneOrMany[str]): Specify a series of tick values to be returned.

        Note:
            - The endpoint /iserver/accounts must be called prior to /iserver/marketdata/snapshot.
            - For derivative contracts, the endpoint /iserver/secdef/search must be called first.
        """
        params = {
            'conids': ','.join(conids),
            'fields': ','.join(fields)
        }
        return self.get(f'iserver/marketdata/snapshot', params)

    def live_marketdata_snapshot_by_symbol(self: 'IbkrClient', queries: StockQueries, fields: OneOrMany[str]) -> dict:
        """
        Get Market Data for the given symbols(s).

        A pre-flight request must be made prior to ever receiving data.

        Parameters:
            queries (List[StockQuery]): A list of StockQuery objects to specify filtering criteria for stocks.
            fields (OneOrMany[str]): Specify a series of tick values to be returned.

        Note:
            - The endpoint /iserver/accounts must be called prior to /iserver/marketdata/snapshot.
            - For derivative contracts, the endpoint /iserver/secdef/search must be called first.
        """
        conids_by_symbol = self.stock_conid_by_symbol(queries).data

        conids = []
        symbols_by_conids = {}
        for symbol, conid in conids_by_symbol.items():
            conids.append(str(conid))
            symbols_by_conids[conid] = symbol

        # There needs to be a pre-flight request for some reason, so we default to calling this twice
        self.receive_brokerage_accounts()
        self.live_marketdata_snapshot(conids, fields=fields)
        entries = self.live_marketdata_snapshot(conids, fields=fields).data

        results = {}
        for entry in entries:
            try:
                result = {}

                for key, value in entry.items():
                    if key not in snapshot_by_id:
                        continue

                    result[snapshot_by_id[key]] = value
                results[entry['conid']] = result
            except Exception as e:  # pragma: no cover
                _LOGGER.exception(f'Error post-processing live market data for {entry}: {str(e)}')

        # reformat the results by symbol instead of conid
        results_by_symbol = {symbols_by_conids[conid]: result for conid, result in results.items()}

        return results_by_symbol

    def regulatory_snapshot(self: 'IbkrClient', conid: str) -> Result:  # pragma: no cover
        """
        Send a request for a regulatory snapshot. This will cost $0.01 USD per request unless you are subscribed to the direct exchange market data already.

        WARNING: Each regulatory snapshot made will incur a fee of $0.01 USD to the account. This applies to both live and paper accounts.

        Parameters:
            conid (str): Provide the contract identifier to retrieve market data for.

        Note:
            - If you are already paying for, or are subscribed to, a specific US Network subscription, your account will not be charged.
            - For stocks, there are individual exchange-specific market data subscriptions necessary to receive streaming quotes.
        """
        return self.get(f'md/regsnapshot', {'conid': conid})

    def marketdata_history_by_conid(
            self: 'IbkrClient',
            conid: str,
            bar: str,
            exchange: str = None,
            period: str = None,
            outside_rth: bool = None,
            start_time: datetime.datetime = None
    ) -> Result:  # pragma: no cover
        """
        Get historical market Data for given conid, length of data is controlled by 'period' and 'bar'.

        Parameters:
            conid (str): Contract identifier for the ticker symbol of interest.
            bar (str): Individual bars of data to be returned. Possible values– 1min, 2min, 3min, 5min, 10min, 15min, 30min, 1h, 2h, 3h, 4h, 8h, 1d, 1w, 1m.
            exchange (str, optional): Returns the exchange you want to receive data from.
            period (str): Overall duration for which data should be returned. Default to 1w. Available time period– {1-30}min, {1-8}h, {1-1000}d, {1-792}w, {1-182}m, {1-15}y.
            outside_rth (bool, optional): Determine if you want data after regular trading hours.
            start_time (datetime.datetime, optional): Starting date of the request duration.

        Note:
            - There's a limit of 5 concurrent requests. Excessive requests will return a 'Too many requests' status 429 response.
        """
        params = params_dict(
            {
                'conid': conid,
                'bar': bar
            },
            optional={
                'exchange': exchange,
                'period': period,
                'outsideRth': outside_rth,
                'startTime': start_time
            },
            preprocessors={
                'startTime': lambda x: x.strftime('%Y%m%d-%H:%M:%S')
            }
        )

        return self.get('iserver/marketdata/history', params)

    def historical_marketdata_beta(
            self: 'IbkrClient',
            conid: str,
            period: str,
            bar: str,
            outside_rth: bool = None,
            start_time: datetime.datetime = None,
            direction: str = None,
            bar_type: str = None,
    ) -> Result:  # pragma: no cover
        """
        Using a direct connection to the market data farm, will provide a list of historical market data for given conid.

        Parameters:
            conid (str): The contract identifier for which data should be requested.
            period (str): The duration for which data should be requested. Available Values: See HMDS Period Units.
            bar (str): The bar size for which bars should be returned. Available Values: See HMDS Bar Sizes.
            outside_rth (bool, optional): Define if data should be returned for trades outside regular trading hours.
            start_time (datetime.datetime, optional): Specify the value from where historical data should be taken. Value Format: UTC; YYYYMMDD-HH:mm:dd. Defaults to the current date and time.
            direction (str, optional): Specify the direction from which market data should be returned. Available Values: -1: time from the start_time to now; 1: time from now to the end of the period. Defaults to 1.
            bar_type (str, optional): Returns valid bar types for which data may be requested. Available Values: Last, Bid, Ask, Midpoint, FeeRate, Inventory. Defaults to Last for Stocks, Options, Futures, and Futures Options.

        Note:
            - The first time a user makes a request to the /hmds/history endpoints will result in a 404 error. This initial request instantiates the historical market data services allowing future requests to return data. Subsequent requests will return data as expected.
        """
        params = params_dict(
            {
                'conid': conid,
                'period': period,
                'bar': bar
            },
            optional={
                'outsideRth': outside_rth,
                'startTime': start_time,
                'direction': direction,
                'barType': bar_type,
            },
            preprocessors={
                'startTime': lambda x: x.strftime('%Y%m%d-%H:%M:%S')
            }
        )

        return self.get('hmds/history', params)

    def marketdata_history_by_symbol(
            self: 'IbkrClient',
            symbol: Union[str, StockQuery],
            bar: str,
            exchange: str = None,
            period: str = None,
            outside_rth: bool = None,
            start_time: datetime.datetime = None,
    ) -> Result:  # pragma: no cover
        """
        Get historical market Data for given symbol, length of data is controlled by 'period' and 'bar'.

        Parameters:
            symbol (Union[str, StockQuery]): StockQuery or str symbol for the ticker of interest.
            bar (str): Individual bars of data to be returned. Possible values– 1min, 2min, 3min, 5min, 10min, 15min, 30min, 1h, 2h, 3h, 4h, 8h, 1d, 1w, 1m.
            exchange (str, optional): Returns the exchange you want to receive data from.
            period (str): Overall duration for which data should be returned. Default to 1w. Available time period– {1-30}min, {1-8}h, {1-1000}d, {1-792}w, {1-182}m, {1-15}y.
            outside_rth (bool, optional): Determine if you want data after regular trading hours.
            start_time (datetime.datetime, optional): Starting date of the request duration.

        """
        conid = str(self.stock_conid_by_symbol(symbol).data[symbol])
        return self.marketdata_history_by_conid(conid, bar, exchange, period, outside_rth, start_time)

    def marketdata_history_by_conids(
            self: 'IbkrClient',
            conids: Union[List[str], Dict[str, str]],
            period: str = "1min",
            bar: str = "1min",
            outside_rth: bool = True,
            start_time: datetime.datetime = None,
            raise_on_error: bool = False,
            run_in_parallel: bool = True,
    ) -> dict:
        """
        An extended version of the marketdata_history_by_conid method.

        For each conid provided, it queries the marketdata history for the specified symbols. The results are then cleaned up and unified. Due to this grouping and post-processing, this method returns data directly without the Result dataclass.

        Parameters:
            conids (OneOrMany[str]): A list of conids to get market data for.
            exchange (str, optional): Returns the exchange you want to receive data from.
            period (str): Overall duration for which data should be returned. Default to 1w. Available time period– {1-30}min, {1-8}h, {1-1000}d, {1-792}w, {1-182}m, {1-15}y.
            bar (str): Individual bars of data to be returned. Possible values– 1min, 2min, 3min, 5min, 10min, 15min, 30min, 1h, 2h, 3h, 4h, 8h, 1d, 1w, 1m.
            outside_rth (bool, optional): Determine if you want data after regular trading hours.
            start_time (datetime.datetime, optional): Starting date of the request duration.
            raise_on_error (bool, optional): If True, raise an exception if an error occurs during the request. Defaults to False.
            run_in_parallel (bool, optional): If True, send requests in parallel to speed up the response. Defaults to True.

        Note:
            - This method returns data directly without the `Result` dataclass.
        """

        if not isinstance(conids, dict):
            # In case conids aren't a symbol->conid dict, generate a dummy conid->conid dict in order to preserve the functionality. In such case, the 'symbol' variable actually is a 'conid'.
            conids = {conid: conid for conid in conids}

        static_params = {"period": period, "bar": bar, "outside_rth": outside_rth, 'start_time': start_time}
        requests = {symbol: {"kwargs": {'conid': conid} | static_params} for symbol, conid in conids.items()}

        # when there is only one conid, we avoid running in parallel
        if run_in_parallel and len(conids) > 1:
            # /iserver/marketdata/history accepts 5 concurrent requests in theory, but sometime throttles above 4
            market_history_responses = execute_in_parallel(self.marketdata_history_by_conid, requests=requests, max_workers=4)
        else:
            market_history_responses = {
                symbol: self.marketdata_history_by_conid(**request['kwargs'])
                for symbol, request in requests.items()
            }

        results = cleanup_market_history_responses(market_history_responses, raise_on_error=raise_on_error)

        return results

    @ensure_list_arg('queries')
    def marketdata_history_by_symbols(
            self: 'IbkrClient',
            queries: StockQueries,
            period: str = "1min",
            bar: str = "1min",
            outside_rth: bool = True,
            start_time: datetime.datetime = None,
            raise_on_error: bool = False,
            run_in_parallel: bool = True,
    ) -> dict:
        """
        An extended version of the marketdata_history_by_conids method.

        For each StockQuery provided, it queries the marketdata history for the specified symbols. The results are then cleaned up and unified. Due to this grouping and post-processing, this method returns data directly without the Result dataclass.

        Parameters:
            queries (List[StockQuery]): A list of StockQuery objects to specify filtering criteria for stocks.
            exchange (str, optional): Returns the exchange you want to receive data from.
            period (str): Overall duration for which data should be returned. Default to 1w. Available time period– {1-30}min, {1-8}h, {1-1000}d, {1-792}w, {1-182}m, {1-15}y.
            bar (str): Individual bars of data to be returned. Possible values– 1min, 2min, 3min, 5min, 10min, 15min, 30min, 1h, 2h, 3h, 4h, 8h, 1d, 1w, 1m.
            outside_rth (bool, optional): Determine if you want data after regular trading hours.
            start_time (datetime.datetime, optional): Starting date of the request duration.
            raise_on_error (bool, optional): If True, raise an exception if an error occurs during the request. Defaults to False.
            run_in_parallel (bool, optional): If True, send requests in parallel to speed up the response. Defaults to True.

        Note:
            - This method returns data directly without the `Result` dataclass.
        """
        conids = self.stock_conid_by_symbol(queries).data
        return self.marketdata_history_by_conids(
            conids=conids,
            period=period,
            bar=bar,
            outside_rth=outside_rth,
            start_time=start_time,
            raise_on_error=raise_on_error,
            run_in_parallel=run_in_parallel,
        )

    @ensure_list_arg('conids')
    def marketdata_unsubscribe(self: 'IbkrClient', conids: OneOrMany[str]) -> List[Result]:
        """
        Cancel market data for given conid(s).

        Parameters:
            conids (OneOrMany[str]): Enter the contract identifier to cancel the market data feed. This can clear all standing market data feeds to invalidate your cache and start fresh.
        """
        # we unsubscribe from all conids simultaneously
        unsubscribe_requests = {conid: {'args': [f'iserver/marketdata/unsubscribe'], 'kwargs': {'params': {'conid': int(conid)}}} for conid in conids}
        results = execute_in_parallel(self.post, unsubscribe_requests)

        for conid, result in results.items():
            if isinstance(result, Exception):
                # 404 means that no such subscription was found in first place, which we ignore
                if isinstance(result, ExternalBrokerError) and result.status_code == 404:
                    continue
                raise result

        return results

    def marketdata_unsubscribe_all(self: 'IbkrClient') -> Result:  # pragma: no cover
        """
        Cancel all market data request(s). To cancel market data for a specific conid, see /iserver/marketdata/{conid}/unsubscribe.
        """
        return self.get(f'iserver/marketdata/unsubscribeall')



from ibind.base.queue_controller import QueueAccessor
from ibind.base.rest_client import Result
from ibind.base.subscription_controller import SubscriptionProcessor
from ibind.client.ibkr_client import IbkrClient
from ibind.client.ibkr_ws_client import IbkrWsClient
from ibind.client.ibkr_ws_client import IbkrWsKey
from ibind.client.ibkr_ws_client import IbkrSubscriptionProcessor
from ibind.client.ibkr_utils import StockQuery, make_order_request, OrderRequest, QuestionType, Answers
from ibind.client.ibkr_definitions import snapshot_keys_to_ids
from ibind.support.errors import ExternalBrokerError
from ibind.support.logs import ibind_logs_initialize
from ibind.support.py_utils import execute_in_parallel, patch_dotenv

__all__ = [
    'ibind_logs_initialize',
    'IbkrClient',
    'IbkrWsClient',
    'IbkrWsKey',
    'IbkrSubscriptionProcessor',
    'SubscriptionProcessor',
    'StockQuery',
    'make_order_request', # deprecated, remove after v0.1.14
    'OrderRequest',
    'QuestionType',
    'Answers',
    'snapshot_keys_to_ids',
    'Result',
    'QueueAccessor',
    'execute_in_parallel',
    'ExternalBrokerError',
]

patch_dotenv()


import time
import unittest
from unittest.mock import MagicMock, patch

from ibind.support.py_utils import ensure_list_arg, execute_in_parallel, execute_with_key, wait_until


class TestEnsureListArgU(unittest.TestCase):

    @ensure_list_arg('arg')
    def sample_function(self, arg):
        return arg

    def test_ensure_list_arg_with_list(self):
        input_arg = [1, 2, 3]
        self.assertEqual(self.sample_function(input_arg), input_arg)

    def test_ensure_list_arg_with_non_list(self):
        input_arg = 1
        self.assertEqual(self.sample_function(input_arg), [input_arg])

    def test_ensure_list_arg_with_keyword_arg_list(self):
        input_arg = [1, 2, 3]
        self.assertEqual(self.sample_function(arg=input_arg), input_arg)

    def test_ensure_list_arg_with_keyword_arg_non_list(self):
        input_arg = 1
        self.assertEqual(self.sample_function(arg=input_arg), [input_arg])

    def test_ensure_list_arg_with_missing_arg(self):
        with self.assertRaises(TypeError):
            self.sample_function()


class TestExecuteInParallelU(unittest.TestCase):

    def _func(self, v1, v2):
        if v1 == 1:
            time.sleep(self.delay)
            return 'result1'
        elif v2 == 2:
            return 'result2'
        else:
            return 'unknown'

    def setUp(self):
        self.delay = 0
        self.func = MagicMock(side_effect=self._func)
        self.func.__name__ = 'TEST_FUNCTION'
        self.requests_dict = {
            'req1': {'args': [1, 0], 'kwargs': {}},
            'req2': {'args': [0], 'kwargs': {'v2': 2}}
        }
        self.requests_list = [{'args': [1, 0], 'kwargs': {}}, {'args': [0], 'kwargs': {'v2': 2}}]

    def test_execute_in_parallel_with_dict(self):
        results = execute_in_parallel(self.func, self.requests_dict)
        self.assertEqual(results, {'req1': 'result1', 'req2': 'result2'})
        self.assertEqual(self.func.call_count, 2)

    def test_execute_in_parallel_with_list(self):
        self.delay = 0.1
        results = execute_in_parallel(self.func, self.requests_list)
        self.assertEqual(results, ['result1', 'result2'])
        self.assertEqual(self.func.call_count, 2)

    def test_execute_with_key_success(self):
        result = execute_with_key('key', self.func, 1, v2=2)
        self.func.assert_called_with(1, v2=2)
        self.assertEqual(result, ('key', 'result1'))

    def test_execute_with_key_exception(self):
        self.func.side_effect = Exception("error")
        result = execute_with_key('key', self.func, 1, v2=2)
        self.assertIsInstance(result[1], Exception)

    def test_execute_in_parallel_rate_limiting(self):
        start_time = time.time()

        # Simulate a slow function to test rate limiting
        def slow_func():
            time.sleep(0.05)
            return "slow_result"

        requests = {i: {'args': [], 'kwargs': {}} for i in range(20)}  # 10 requests
        max_per_second = 10  # Limit to 5 requests per second
        results = execute_in_parallel(slow_func, requests, max_per_second=max_per_second)

        duration = time.time() - start_time
        self.assertGreaterEqual(duration, 1.05)  # Should take at least 1.1 seconds to complete all requests
        self.assertEqual(len(results), 20)


class TestWaitUntilU(unittest.TestCase):

    def test_wait_until_condition_met(self):
        condition = MagicMock(return_value=True)
        self.assertTrue(wait_until(condition))
        condition.assert_called()

    def test_wait_until_condition_not_met(self):
        condition = MagicMock(return_value=False)
        self.assertFalse(wait_until(condition, timeout=0.1))
        condition.assert_called()

    @patch('ibind.support.py_utils._LOGGER.error')
    def test_wait_until_timeout_message(self, mock_logger_error):
        condition = MagicMock(return_value=False)
        timeout_message = "Condition not met within timeout"
        self.assertFalse(wait_until(condition, timeout_message=timeout_message, timeout=0.1))
        mock_logger_error.assert_called_with(timeout_message)

    def test_wait_until_timeout(self):
        start_time = time.time()
        condition = MagicMock(return_value=False)
        timeout = 0.1
        self.assertFalse(wait_until(condition, timeout=timeout))
        duration = time.time() - start_time
        self.assertAlmostEqual(duration, timeout, delta=0.02)




from typing import Callable, Optional
from unittest.mock import MagicMock

def init_wsa_mock(
        wsa_mock: MagicMock,
        url: str,
        on_open: Callable = None,
        on_message: Callable = None,
        on_error: Callable = None,
        on_close: Callable = None,
        cookie:str=None,
        header:dict=None,
):
    wsa_mock.url = url
    wsa_mock.cookie = cookie
    wsa_mock.header = header

    wsa_mock.__on_open__ = on_open
    wsa_mock.__on_message__ = on_message
    wsa_mock.__on_error__ = on_error
    wsa_mock.__on_close__ = on_close

    wsa_mock._on_open.side_effect = on_open
    wsa_mock._on_message.side_effect = on_message
    wsa_mock._on_error.side_effect = on_error
    wsa_mock._on_close.side_effect = on_close

    wsa_mock.last_ping_tm = 0
    wsa_mock.keep_running = False

    return wsa_mock


def send(wsa_mock: MagicMock, message: str):
    wsa_mock._on_message(wsa_mock, message)


def close(wsa_mock: MagicMock, status:str=None):
    wsa_mock.keep_running = False
    wsa_mock._on_close(wsa_mock, None, None)


def run_forever(wsa_mock: MagicMock, sslopt: dict = None, ping_interval: float = 0, ping_timeout: Optional[float] = None):
    wsa_mock.keep_running = True
    wsa_mock._on_open(wsa_mock)

def create_wsa_mock():
    wsa_mock = MagicMock()

    wsa_mock.send.side_effect = lambda *args, **kwargs: send(wsa_mock, *args, **kwargs)
    wsa_mock.close.side_effect = lambda status=None: close(wsa_mock, status)
    wsa_mock.run_forever.side_effect = lambda *args, **kwargs: run_forever(wsa_mock, *args, **kwargs)

    return wsa_mock
